/*
pyodide-mkdocs-theme
Copyleft GNU GPLv3 🄯 2024 Frédéric Zinelli

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.
If not, see <https://www.gnu.org/licenses/>.
*/


import { jsLogger } from 'jsLogger'
import {
  escapePyodideCodeForJsTemplates,
  PythonError,
  toSnake,
} from 'functools'
import {
  getFullStdIO,
  pyodideFeatureRunCode,
  setupStdIO,
} from '0-generic-python-snippets-pyodide'
import { generateErrorLog } from '1-error_logs-pyodide'
import { installPythonPackages } from '1-packagesInstaller-install-pyodide'




/**Build a "context" object, holding the various state information for the current
 * "sub-action/step" in the current overall user action (RuntimeManager).
 * (creates a bare object, without methods)
 * */
export class Ctx {

  static baseCtx(){
    return {
      success: true,                // false on errors only, not if just "skipped"

      err: null,                    // Error instance
      stdErr: "",                   // Error as string (msg to output, formatted)
      isAssertErr:false,
      stdout: "",
      qualname: "<unknown>",        // "qualified name" of the section + method to run in the current action (debugging purpose)

      keepRunningOnAssert: false,   // If true, consider the section ran successfully on AssertionError (this is env related)
      isEnvSection: true,           // Any of env, env_term, ... => "not user code or cmd"
      applyExclusionsIfAny: false,  // Apply exclusions specifically on this run (if any). Note: `!isEnvSection` is not specific enough!
      kwsExclusions: undefined,     // Define if some exclusions logic has to be applied or not on the way
      methodsExclusions: undefined, // Define if some exclusions logic has to be applied or not on the way
      runtimeExclusions: undefined, // Define if some exclusions logic has to be applied or not on the way
      runtimeExclusionSetupSuccessful: false,  // flag
      autoImport: true,             // If true, apply automatically install/imports logistic
      section: "<unknown>",         // as in "PYODIDE:{section}", or "unknown" instead. Visible in the stacktrace on errors.
      archiveSuccess: false,        // If true, the section will be registered as ran successfully (if it is so). To use on the last action of the section.
      method: ()=>null,             // Runner method to... run when everything is ready
      code: ".",                    // If empty string, the Runner method won't ever be called. Default is not empty, because some Runner won't need to pass `code` to the ctx to know what to do (or the code needed is not known yet: Features)
      methodArgs: [],               // Additional arguments to pass to the Runner method
      logConfig: {},                // Config for generateErrorLog
    }
  }


  static with(ctx, runtime){
    const baseCtx = Ctx.baseCtx()
    let section   = ctx && ctx.section || baseCtx.section

    if(!ctx) ctx={}                     // no argument

    else if(typeof(ctx)=='string'){     // Environment section name only
      section = ctx
      ctx = {
        archiveSuccess:      true,      // always, with environment codes
        code:                runtime.runner[`${ section }Content`].trim(),
        keepRunningOnAssert: section.startsWith('env'),
        method:              (ctx)=>runtime.installImportsAndRunEnvCode(ctx),
                             // Avoids the wrong binding to runtime.runner
      }
    }

    ctx={
      ...baseCtx,
      qualname: `${ section || '?' }_${ ctx.method.name }`,
      ...ctx,
      section,
    }
    const needExclusions  = !ctx.isEnvSection && ctx.applyExclusionsIfAny
    ctx.kwsExclusions     ??= needExclusions && runtime.excludedKws.length>0
    ctx.methodsExclusions ??= needExclusions && runtime.excludedMethods.length>0
    ctx.runtimeExclusions ??= needExclusions && (runtime.excluded.length>0 || runtime.recLimit > 0)

    return ctx
  }
}









/**Manage the environment when in need to execute some code in pyodide.
 *
 * Holds all the logic to manage all the possible outcomes:
 * Used as "throw away" data structure, once the full user action is done: this allows
 * to mutate "everything you need", being sure it wont affect the original data in the
 * `PythonSectionsRunner` instance.
 * */
export class RuntimeManager {


  get excluded()        { return this.runner.excluded }
  get excludedMethods() { return this.runner.excludedMethods }
  get excludedKws()     { return this.runner.excludedKws }
  get recLimit()        { return this.runner.recLimit }
  get whiteList()       { return this.runner.whiteList }



  constructor(runner){
    this.runner = runner    // PyodideSectionsRunner, or one of its parent class/object

    // --- Python logistic related (may be mutated on the way) ---
    this.autoLogAssert   = null     // default for the PUBLIC tests...
    this.purgeStackTrace = null     // default for the PUBLIC tests...
    this.withStdOut      = null     // default for the PUBLIC tests...
    this.refreshStateWith()         // Apply actual defaults (avoiding defaults duplication)

    // --- Runtime logistic related (may be mutated on the way) ---
    this.stdErr       = ""          // First encountered error message
    this.isAssertErr  = false       // Is the error an assertion error
    this.gotBigFail   = false       // If true, new errors won't replace the current one
    this.finalMsg     = ""          // Potential extra message to display at the very end, if truthy.
                                    //  (will be printed even if this.stopped is true)

    this.ran = {start: true}        // Keep track of the successfully run env sections (or raising assertion errors)
    this.dependencies = {
      env:      'start',
      envTerm:  'env',
      code:     'env',              // (actually not used as a condition for now)
      cmd:      'envTerm',          // (actually not used as a condition for now)
      postTerm: 'envTerm',
      post:     'env',
    }
  }

  get stopped(){ return Boolean(this.stdErr) }


  /**Update the state of the current RuntimeManager with the given data, and use
   * default values for missing fields.
   * */
  refreshStateWith(update={}){
    update = {
      autoLogAssert:   true,
      purgeStackTrace: false,
      withStdOut:      true,
      ...update
    }
    for(const prop in update){
      this[prop] = update[prop]
    }
  }


  /**Update the sections dependency graph, to modify the conditions to run or not a section.
   * */
  changeDependency(source, target){
    if(this.dependencies[source]===undefined){
      throw new Error("Invalid source dependency: " + source)
    }
    this.dependencies[source] = target
  }

  cleanup(){
    return this.runner = this.runCodeAsync = null   // Facilitate GC...
  }



  stillRunnable(ctx){
    let outcome = ctx.isEnvSection
                  ? this.ran[ this.dependencies[ctx.section] ]
                  : !this.stopped
    outcome &&= !this.gotBigFail
    return outcome
  }



  /**Generic, main "action" runner for environment code.
   * */
  async installImportsAndRunEnvCode(ctx){
    if(!ctx.code) return;
    await pyodide.runPythonAsync(ctx.code, {filename: `<${ toSnake(ctx.section) }>`})
  }


  /**Run one Runner method (async), surrounding it with all necessary logics, like:
   *
   *    - setup+teardown stdout
   *    - setup+teardown exclusions
   *    - give feedback
   *    - try/catch where appropriate (handling properly the exclusions removal case)
   *
   * This method should never actually throw, except in "BigFail" cases.
   * Note: NOT this.stdErr, which is more global, while the current run could involve `post`
   *       actions that should potentially occur even if an error has already occur previously
   *       in the current user's action.
   *
   * It uses a ctx object, which holds the local executions data, and will merge the data
   * when appropriate in the RunnerManager (which holds runtime information for the current
   * user's action).
   *
   * Inside this "runWithCtx" call, subsequent methods calls are skipped if an error has been
   * registered in the @ctx object, unless the method is specified as "should `always` happen"
   * */
  async runWithCtx(ctx){
    ctx = Ctx.with(ctx, this)

    if(!this.stillRunnable(ctx)){
      jsLogger('[Runtime] - SKIPPED', ctx.qualname)
      return ctx
    }

    jsLogger('[Runtime] - running ', ctx.qualname)
    // No need for error extra handling: big fails handled in lockedRunnerWithBigFailWarningFactory

    this.runner.allowPrint = this.withStdOut
    const astExclusions    = ctx.kwsExclusions || ctx.methodsExclusions

                              await this._runCaught(ctx, setupStdIO)
    if(astExclusions)         await this._runCaught(ctx, this.setupAstExclusions)
    if(ctx.autoImport)        await this._runCaught(ctx, this.installOrImportModules) // BEFORE runtime exclusions!
    if(ctx.runtimeExclusions) await this._runCaught(ctx, this.setupRuntimeExclusions)
    if(!ctx.isEnvSection)     await this._runCaught(ctx, this.clearAutoRun)
                              await this._runCaught(ctx, this.applyRunnerMethod)
    if(ctx.runtimeExclusions) await this._runCaught(ctx, this.removeExclusions, {always:true})
                              await this._runCaught(ctx, this.teardownManager, {always:true, reThrow:true})

    this.runner.allowPrint = true
    this.runner.giveFeedback(ctx.stdErr)

    if(ctx.archiveSuccess){
      this.ran[ctx.section] = ctx.success
    }
    jsLogger('[Runtime] - done', ctx.qualname)
    return ctx
  }




  async _runCaught(ctx, method, conf={}){
    jsLogger('[Runtime] - _runCaught ', ctx.qualname, method.name)
    if(ctx.err && !conf.always) return
    try{
      await method.call(this, ctx, conf)
    }catch(e){
      jsLogger('[Runtime] - ERROR ', ctx.qualname)
      if(CONFIG.loggerOptions.ACTIVATE){
        console.log('[jsLogger activated, causing the following log:')
        console.error(e)
      }
      ctx.err = e
      if(conf.rethrow) throw e
    }
  }


  async setupAstExclusions(ctx){
    jsLogger('[CheckPoint] - running - setup keywords exclusions')
    if(ctx.code){
      astExclusions(ctx.code, this.excludedKws, this.excludedMethods)
    }
  }


  // async setupMethodsExclusions(ctx){
  //   jsLogger('[CheckPoint] - running - setup methods exclusions')
  //
  //   const calls = new Set()
  //   for(const m of ctx.code.matchAll(/[.][a-z]\w*/gi)){
  //     calls.add(m[0])
  //   }
  //   const nope = this.excludedMethods.filter(methodCall=>calls.has(methodCall))
  //
  //   if(nope.length){
  //     const plural = nope.length>1 ? "s":""
  //     const nopes  = nope.map( s=>s.slice(1) ).join(', ')
  //     const msg    = `${ CONFIG.MSG.exclusionMarker } method${plural}: ${ nopes }`
  //     throw new PythonError(msg)
  //   }
  // }


  async setupRuntimeExclusions(ctx){
    jsLogger('[CheckPoint] - running - setup runtime exclusions')
    setupExclusions(this.excluded, this.recLimit)
    ctx.runtimeExclusionSetupSuccessful = true
  }


  async clearAutoRun(ctx){
    pyodideFeatureRunCode('autoRunCleaner')
  }


  async installOrImportModules(ctx){
    await installPythonPackages(this.runner, ctx.code, this, ctx.isEnvSection)
  }

  async removeExclusions(ctx){
    jsLogger('[CheckPoint] - running - removing exclusions')
    if(ctx.runtimeExclusionSetupSuccessful){
      restoreOriginalFunctions(this.excluded)
    }
  }


  async teardownManager(ctx){
    getFullStdIO()
    this.handleError(ctx)
  }




  /**Run the desired Runner method, once all the setup has been managed.
   * If it succeeds (aka, the method doesn't throw any error), the ctx is marked as successful.
   *
   * The Runner method is supposed to take this kind of signature:
   *
   *      method(...ctx.methodArgs, ctx, runtime)
   *
   * Where `ctx.methodArgs` are arguments the RuntimeManager cannot know about in the first place.
   * The two last arguments are always passed but may be ignored on the actual method signature, if
   * it doesn't make use of them.
   * */
  async applyRunnerMethod(ctx){
    jsLogger('[CheckPoint] - running method runner -', ctx.qualname)
    await ctx.method.call(this.runner, ...ctx.methodArgs, ctx, this) // always send extras args (in case useful)
    ctx.success = true
  }


  /**In case an error has been registered in @ctx
   * */
  handleError(ctx, replLogConf=null){
    if(!ctx.err) return;

    this.finalMsg = ""

    ;[ctx.stdErr, ctx.isAssertErr] = generateErrorLog(ctx.err, replLogConf || ctx.logConfig)

    // If ever multiple errors happen, an assertion error will always be the very
    // first one, so always keep if any (avoid the need to condition the update)
    if(!this.gotBigFail){
      this.isAssertErr ||= ctx.isAssertErr
      this.stdErr        = ctx.stdErr
    }
    ctx.success = ctx.keepRunningOnAssert && ctx.isAssertErr
  }
}

















/*
------------------------------------------------------------------
                      Manage code exclusions
------------------------------------------------------------------
*/


/**Put in place code exclusions. Are handled:
 *   - builtin function calls
 *   - imports
 *   - method calls (done through a simple string check in the code, in
 *     runPythonCodeWithOptionsIfNoStdErr)
 *
 *
 * ## RATIONALS:
 *
 * To forbid the use of some functions or packages, replace them in the global scope by "functions"
 * that look "more or less the same", but will raise an error when called, or when used in the
 * wrong way (imports).
 *
 *
 * ## PROBLEMS & CONTEXT:
 *
 * 1. Pyodide itself uses various functions to run python code:
 *      - eval is used in pyodide.runPython
 *      - reversed and/or min and/or max may be used when building a stacktrace when an error is
 *        thrown in python
 * 2. This forbids to replace the __builtins__ versions of those functions (see about imports)...
 * 3. ...but the __main__ script is run separately of pyodide actual "python runtime".
 *
 *
 * ## SOLUTION FOR BUILTINS FUNCTIONS:
 *
 * - Redeclare forbidden things in the global scope, through `globals()`, using an object that
 *   will systematically throw an ExclusionError when it's called.
 * - Since those are in the global scope, they are visible through `dir()`, so add some make up
 *   to them, using a class that redefines its __qualname__ and __repr__, so that they are less
 *   obvious as "anti-cheats" (it will still remain obvious for those who know enough. But if they
 *   can find about that, they probably could solve the problem the right way anyway...).
 * - Pyodide runtime won't see those globals, so it is not affected in any way, only the user's
 *   and tester's codes are.
 * - The (hidden) function `__move_forward__('builtin_name')` (see documentation) can be used in
 *   the tests to get back the original builtin. If used, it must be done inside a closure, so
 *   that the original builtin doesn't override the "Raiser" in the global scope (see below).
 * - Since the hacked versions are available to the user in the global runtime, they could just
 *   delete them to get back the access to the original builtins. To limit this risk, an extra
 *   check is done after the user's code and tests has been run, verifying that the hacked
 *   functions are still defined in the global scope, and that they still are the original
 *   Raiser objects.
 *
 *
 * ## SOLUTION FOR IMPORTS
 *
 * The problem with `import` is that they actually go directly through `__builtins__.__import__`.
 * So in that case, there is no other choice than hacking directly the builtin, and then put it
 * back in place when not useful anymore.
 *
 *
 * ## RECURSION LIMIT
 *
 * The sys module function is directly hacked, then put back in place: meaning, the function
 * setrecursionlimit is also replaced at user's runtime with a Raiser object.
 *
 * */
const setupExclusions =(excluded, recLimit)=>{
  // Store None in the __builtins___ dict for things that aren't builtin functions, aka, names
  // of forbidden module.

  /** WARNING!
   *  Keep in mind that the code of the Raiser instances will run "in context".
   *  This means it will be subject to existing exclusions, so it must never use a function that
   *  could be forbidden. Possibly...
   *  For this reason, copies of all the builtins used in the Raiser code are stored locally, to
   *  be sure the Raiser won't use Raiser instances... XD
   * */
  const code = `
  @__builtins__.auto_run
  def _hack_exclusions_setup():

      class Raiser:
          __name__ = __qualname__ = 'function'

          def __init__(self, key):  self.key = key

          def __repr__(self): return f"<built-in function {self.key}>"

          def __call__(self, *a, **kw):
              key = self.key

              head = a and base_isinstance(a[0],base_str) and a[0].split(".")[0]

              is_forbidden = (
                  key != '__import__' or
                  key == '__import__' and head in dct
              )
              if is_forbidden:
                  that = key if key!='__import__' else head
                  ExclusionError.throw(that)

              # if reaching this point, the call is a valid import, so apply it:
              return base_import(*a,**kw)


      # Store the originals used here to avoid troubles with exclusions at runtime:
      base_import = __import__
      base_str = str
      base_isinstance = isinstance


      dct = __builtins__.__builtins___ = {}
      raiser_import = Raiser('__import__')
      dct['__import__'] = [base_import, raiser_import]
      __builtins__.__import__ = raiser_import


      glob_dct = globals()
      exclusions = ${ JSON.stringify(excluded) }
      for key in exclusions:
          stuff = getattr(__builtins__, key, None)
          dct[key] = [stuff, None]
          # => the dict will store [None,None] for module names

      if ${ recLimit } != -1:
          import sys
          sys.setrecursionlimit(${ recLimit })
          dct['setrecursionlimit'] = [sys.setrecursionlimit, None]

      for key,lst in dct.items():
          stuff = lst[0]
          if callable(stuff) and key!='__import__':       # import has already been handled
              # Store the reference of the raiser in lst, to check against it later:
              glob_dct[key] = lst[1] = Raiser(key)

      # auto_run added for verification purpose only, but it must stay usable:
      dct['auto_run'] = auto_run
`
  pyodide.runPython(code)
}






/**Cancel the code exclusions (done as soon as possible, to restore pyodide's normal behaviors).
* */
const restoreOriginalFunctions = exclusions =>{

  // Don't use auto_run, in case someone messes with it...
  const code = `
def _hack_unexclude():
  dct = __builtins__.__builtins___
  G = globals() if "globals" not in dct else dct['globals']()

  # Handle special behavior of auto_run:
  key = "auto_run"
  auto_run = dct.pop(key)
  bad_auto_run = G.get(key, __builtins__.auto_run) is not auto_run
  __builtins__.auto_run = auto_run

  # Restore everything before raising anything:
  not_ok = []
  for key,(func,raiser) in dct.items():

      if key == '__import__':
          __builtins__.__import__ = func

      else:
          if raiser is not None and raiser is not G.get(key):
              not_ok.append(key)
          if func is not None:
              if key in G:
                  del G[key]  # "Unshadow" the builtin
              if key == 'setrecursionlimit':
                  func(1000)

  if bad_auto_run:
      ExclusionError.throw("the auto_run tool...")
  if not_ok:
      ExclusionError.throw("${ exclusions }")

_hack_unexclude()
del _hack_unexclude
`
  pyodide.runPython(code)
}







/**Attempt to "optimize" the NodeVisitor class to not lose time on attribute accesses when
 * none are to check, or on keywords check when none are to check.
 * */
const astExclusions =(code, excludedKws, excludedMethods)=> {

  const attrChecker = !excludedMethods.length ? "" : `
        def visit_Attribute(self, node:ast.Attribute):
            if node.attr in forbidden_attrs:
                wrongs.add('.'+node.attr)
            super().generic_visit(node)
`
  const kwsChecker = !excludedKws.length ? "":`
        def generic_visit(self, node):
            name = node.__class__.__name__
            if name in classes_dct:
                self.check_those(*classes_dct[name], gather_all=True)
            super().generic_visit(node)


        def check_those(self, *stuff:str, gather_all=False):
            if gather_all:
                for kw in stuff:
                    if kw in forbidden_kws:
                        wrongs.add( kw.replace('_',' ') )
            else:
                for kw in stuff:
                    if kw in forbidden_kws:
                        wrongs.add( kw.replace('_',' ') )
                        break


        #----------------------------------------------


        def visit_Constant(self, node:ast.Constant):
            self.check_those( repr(node.value) )


        def visit_alias(self, node:ast.Constant):
            if node.asname:
                self.check_those('as')
            super().generic_visit(node)


        # if, else (no loops')
        def visit_IfExp(self, node:ast.If):
            self.visit_If(node, 'if', 'if')

        # if, elif, else (no loops')
        def visit_If(self, node:ast.If, kw='if', next_in_line='elif'):
            if kw in forbidden_kws:
                wrongs.add(kw)
            self.visit(node.test)
            for child in node.body: self.visit(child)
            for child in node.orelse:
                if isinstance(child, ast.If):
                    self.visit_If(child, next_in_line, next_in_line)
                elif 'else' in forbidden_kws:
                    wrongs.add('else')
                    self.visit(child)


        # from (in raise statements)
        def visit_Raise(self, node:ast.Raise):
            self.check_those('raise')
            if node.cause:
                self.check_those('from', 'raise_from')
            super().generic_visit(node)

        def visit_ExceptHandler(self, node:ast.ExceptHandler):
            to_check = ('except',)
            if node.name:
                to_check += ('as', 'except_as')
            self.check_those(*to_check, gather_all=True)
            super().generic_visit(node)

        def visit_Try(self, node):
            to_check = ('try',)
            if node.orelse:
                to_check += ('else', 'try_else')
            if node.finalbody:
                to_check += ('finally',)
            self.check_those(*to_check, gather_all=True)
            super().generic_visit(node)

        def visit_TryStar(self, node):    # 3.11+ only
            self.visit_Try(node)


        def visit_loop(self, node, name:str, stuff:str, is_comp, is_async):
            to_check = (stuff,)
            if stuff=='for':
                to_check += ('for_comp',) if is_comp else ('for_inline',)
            if not is_comp and node.orelse:
                to_check += ('else', stuff+'_else')
            if is_comp and node.ifs:
                to_check += ('if',)
            if is_async:
                to_check += ('async', 'async_'+stuff)
            self.check_those(*to_check, gather_all=True)
            super().generic_visit(node)

        def visit_For(self, node):            self.visit_loop(node, 'For',           'for',   False, False)
        def visit_While(self, node):          self.visit_loop(node, 'While',         'while', False, False)
        def visit_AsyncFor(self, node):       self.visit_loop(node, 'AsyncFor',      'for',   False, True)
        def visit_comprehension(self, node):  self.visit_loop(node, 'comprehension', 'for',   True,  node.is_async)


        def visit_async_stuff(self, node, kind:str):
            self.check_those(kind, 'async_'+kind)
            super().generic_visit(node)

        def visit_AsyncFunctionDef(self, node): self.visit_async_stuff(node, 'def')
        def visit_AsyncWith(self, node):        self.visit_async_stuff(node, 'with')
`

  const checkCode = `
@auto_run
def _hack_check_excluded_keywords():
    import ast, js

    try:
        tree = ast.parse("""${ escapePyodideCodeForJsTemplates(code) }""")
    except:
        return  # invalid syntax => nothing to check


    classes_dct     = js.config().PYTHON_KEYWORDS.to_py()
    forbidden_kws   = {${ JSON.stringify(excludedKws).slice(1,-1) }}
    forbidden_attrs = {${ JSON.stringify(excludedMethods).slice(1,-1) }}
    wrongs          = set()

    class ExclusionsVisitor(ast.NodeVisitor):
        ${ attrChecker }
        ${ kwsChecker }

    ExclusionsVisitor().visit(tree)
    if wrongs:
        to_clean_up = {'f str', 'f string'}
        if to_clean_up & wrongs:
            wrongs -= to_clean_up
            wrongs.add('f-string')
        raise ExclusionError.throw( ', '.join(sorted(wrongs, key=str.casefold)) )
`
    pyodide.runPython(checkCode)
}
